"""UI for SpikeInterface extracellular data processing pipelines.

The application enables a user to constuct a pipeline consisting of elements
associated with extracellular data recording extraction, pre-processing,
sorting, and curation.  The user can configure the properties of each
element, and once satisifed with the pipeline construction and element
configuration operate the pipeline.

Modules:
    ce_view.py - UI to configure element parameters
    cfg.py - Constants and globals used throughout
    cp_view.py - UI to select, insert, and manipulate elements in pipeline
    el_model.py - SpikeElementModel bridges SpikeElement to/from UI
    op_view.py - UI to manage pipeline commands (e.g., clear, run, queue)
    pi_model.py - SpikePipeline model bridges element pipeline to/from UI
    spike_element.py - SpikeElement class from which to subclass elements
    spikely.py - Main application sets up UI and executes event loop
"""

import sys

import PyQt5.QtWidgets as qw
import PyQt5.QtGui as qg

from spikely.op_view import OperatePipelineView
from spikely.cp_view import ConstructPipelineView
from spikely.ce_view import ConfigureElementView
from spikely.pi_model import SpikePipelineModel
from spikely.el_model import SpikeElementModel

import pkg_resources

from . import config as cfg
from spikely.version import __version__


class SpikelyMainWindow(qw.QMainWindow):
    """Main window for the application"""

    def __init__(self):

        super().__init__()

        # Active pipeline and element models
        self._element_model = SpikeElementModel()
        self._pipeline_model = SpikePipelineModel(
            self._element_model)

        # Enhances print() use for debug
        sys.stdout.flush()

        self._init_ui()

    def _init_ui(self):
        """Assembles the main UI from delegated sub-views."""

        # Application main window setup
        self.setWindowTitle("Spikely")
        self.setGeometry(100, 100, 1152, 448)

        spikely_png_path = pkg_resources.resource_filename(
            'spikely.resources', 'spikely.png')

        self.setWindowIcon(qg.QIcon(spikely_png_path))
        self.statusBar().addPermanentWidget(
            qw.QLabel("Version " + __version__))

        # Core application UI in main_frame as CentralWidget of QMainWindow
        main_frame = qw.QFrame()
        self.setCentralWidget(main_frame)
        main_frame.setLayout(qw.QVBoxLayout())

        """ Lay out Construction Pipeline (cp) and Configure Element (ce)
        views in a frame at top of main window from left to right
        """
        cp_ce_splitter = qw.QSplitter()
        cp_ce_splitter.setChildrenCollapsible(False)

        # Actual widget construction done in View classes
        cp_ce_splitter.addWidget(ConstructPipelineView(
            self._pipeline_model, self._element_model))
        cp_ce_splitter.addWidget(ConfigureElementView(
            self._pipeline_model, self._element_model))
        cp_ce_splitter.setSizes([328, 640])
        main_frame.layout().addWidget(cp_ce_splitter)

        # Lay out Operate Pipeline (op) view at bottom of main window
        main_frame.layout().addWidget(OperatePipelineView(
            self._pipeline_model, self._element_model))

        main_frame.layout().addStretch(1)  # Pushes app window widgets up

        # Allows any module to post a status message to main window
        # cfg.status_bar = self.statusBar()


def launch_spikely():
    app = qw.QApplication(sys.argv)
    cfg.main_window = SpikelyMainWindow()
    cfg.status_bar = cfg.main_window.statusBar()
    cfg.main_window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    launch_spikely()



import spikeextractors as se
import spiketoolkit as st

################################ Extractors ###################################

# Create a toy example dataset (returns a recording extractor and sorting extractor)
recording, sorting_true= se.example_datasets.toy_example(duration=60)
print("Recording Extractor: " + str(recording))
print("Sorting Extractor: " + str(sorting_true) + '\n')

############################### Preprocessing #################################

# Filter recording extractor for the timeseries data (acts as a recording extractor that also filters!)
filter_recording = st.preprocessing.BandpassFilterRecording(recording, freq_min=300, freq_max=6000)
print("Filter Recording Extractor: " + str(filter_recording) + '\n')

################################### Sorting ###################################

#We have a string name for each sorter (we use that to get the class)
sorter_name = 'klusta'
output_folder='tmp_klusta'

#Fetch the sorter class corresponding to the sorter_name
SorterClass = st.sorters.sorter_dict[sorter_name]

#Instatiating the sorter class is generic (we always pass in these parameters)
sorter = SorterClass(recording=recording, output_folder=output_folder, grouping_property=None, parallel=False,
                     debug=False, delete_output_folder=False)
print("Sorter Class: " + str(sorter) + '\n')
#Get default params for specific sorter
params = SorterClass.default_params()

#set params for the sorter using default params
sorter.set_params(**params)
print("Default Params: "+ str(sorter.params) + '\n')

#Can update params by passing in the user entries
sorter.set_params(threshold_strong_std_factor=6, threshold_weak_std_factor=1)

#Run the sorter and save to the output file
sorter.run()

#Get a sorting object result from the output file
sorting_ms4 = sorter.get_result()
print("Sorting Extractor for ms4: " + str(sorting_ms4) + '\n')

################################ CURATION ###############################

# This is function call from the curation module, doesn't return anything (we support this)
st.postprocessing.export_to_phy(recording, sorting_ms4, output_folder='phy')
print("")

#Returns a sorting extractor object which has some curated results inside
curated_sorting = st.curation.ThresholdMinNumSpike(sorting_ms4, min_num_spike_threshold=120)
print("Curated Sorting Extractor: " + str(curated_sorting) + '\n')

#The curation gets rid of all units with less than 50 spikes
print("Original Units: " + str(sorting_ms4.get_unit_ids()))
print("Curated Units: " + str(curated_sorting.get_unit_ids()))



from .spike_element import SpikeElement
import spikeextractors as se
import os


class Curator(SpikeElement):
    """Curator class"""

    def __init__(self, interface_class, interface_id):
        SpikeElement.__init__(self, interface_id, interface_class,
                              interface_class.curator_name)

    def run(self, input_payload, next_element):
        params_dict = {}
        params_dict['sorting'] = input_payload[0]
        output_folder_path = input_payload[1]

        params = self._params
        for param in params:
            param_name = param['name']
            # param_type = param['type']
            # param_title = param['title']
            param_value = param['value']
            params_dict[param_name] = param_value
        sorting = self._interface_class(**params_dict)
        if(next_element is None):
            curated_output_folder = output_folder_path/'curated_phy_results'
            '''
            curated_output_folder =
                Path(curated_output_folder_path).absolute()
            '''
            if not curated_output_folder.is_dir():
                os.makedirs(str(curated_output_folder))
            print("Saving curated results....")
            se.PhySortingExtractor.write_sorting(
                sorting, curated_output_folder)
            print("Done!")
        return sorting, output_folder_path



"""Constants and gloabals used by other spikely modules."""

# Application status bar for user messages.  Set in spikely.py
# Usage: config.status_bar.showMessage('message', config.TIMEOUT)
status_bar = None

# Duration in milliseconds of timeout for temporary status messages
STATUS_MSG_TIMEOUT = 3500

# Application main window used to anchor dialog boxes.  Set in spikely.py
main_window = None

# Used to represent types of elements
EXTRACTOR, PRE_PROCESSOR, SORTER, CURATOR = range(4)

# Identifier to get element object from pipeline model data()
ELEMENT_ROLE = 0x100

# Column IDs used by QTableView to display element parameter data
PARAM_COL = 0
VTYPE_COL = 1
VALUE_COL = 2



"""Creates an MVC view-control for operations on the active pipeline model.

The Operate Pipeline view/control consists of widgets responsible for
user control over the SpikeInterface pipeline of elements used for
extracellular data processing.  Those operations include running, clearing,
and queueing the active pipeline.
"""

import PyQt5.QtWidgets as qw
from . import config as cfg


class OperatePipelineView(qw.QGroupBox):
    """GroupBox of widgets capable of operating the active pipeline.

    Invokes methods on the active pipeline. No public methods other than
    constructor.  All other activites of object are triggered by user
    interaction with sub widgets.
    """

    def __init__(self, pipeline_model, element_model):
        """Initialize parent, set object members, construct UI."""
        super().__init__("Operate Pipeline")  # Applies group box label
        self._pipeline_model = pipeline_model
        self._element_model = element_model

        self._init_ui()

    def _init_ui(self):

        '''
        self._pipeline_model.rowsInserted.connect(self._pipeline_changed)
        self._pipeline_model.rowsRemoved.connect(self._pipeline_changed)
        self._pipeline_model.modelReset.connect(self._pipeline_changed)
        '''

        self.setLayout(qw.QHBoxLayout())

        # Pipeline operation commands
        self._run_btn = qw.QPushButton("Run")
        # self._run_btn.setEnabled(False)
        self._run_btn.clicked.connect(self._run_clicked)
        self.layout().addWidget(self._run_btn)

        self._clear_btn = qw.QPushButton("Clear")
        # self._clear_btn.setEnabled(False)
        self.layout().addWidget(self._clear_btn)

        def clear_clicked():
            self._pipeline_model.clear()
            # Should element model be responsible for this?
            self._element_model.element = None
        self._clear_btn.clicked.connect(clear_clicked)

        self._queue_btn = qw.QPushButton("Queue")
        self._queue_btn.clicked.connect(self._queue_clicked)
        self.layout().addWidget(self._queue_btn)

    def _queue_clicked(self):
        # Pipeline model should be responsible for this
        cfg.status_bar.showMessage(
            "Queue not implemented", cfg.STATUS_MSG_TIMEOUT)

    def _run_clicked(self):
        # Pipeline model should be responsible for this
        self._pipeline_model.run()

    '''
    def _pipeline_changed(self, parent=None, first=None, last=None):
        enabled = self._pipeline_model.rowCount(None) > 0
        self._run_btn.setEnabled(enabled)
        self._clear_btn.setEnabled(enabled)
        self._queue_btn.setEnabled(enabled)
    '''



from .spike_element import SpikeElement


class Sorter(SpikeElement):
    """Preprocessor class"""

    def __init__(self, interface_class, interface_id):
        SpikeElement.__init__(self, interface_id, interface_class,
                              interface_class.sorter_name)

    def run(self, input_payload, next_element):
        base_sorter_param_dict = {}
        base_sorter_param_dict['recording'] = input_payload

        params = self._params
        output_folder = params[0]
        parallel = params[1]
        base_sorter_param_dict[output_folder['name']] = output_folder['value']
        base_sorter_param_dict[parallel['name']] = parallel['value']
        sorter = self._interface_class(**base_sorter_param_dict)
        print(sorter.output_folders)

        sub_sorter_param_dict = {}
        for param in params[2:]:
            param_name = param['name']
            # param_type = param['type']
            # param_title = param['title']
            param_value = param['value']
            sub_sorter_param_dict[param_name] = param_value
        sorter.set_params(**sub_sorter_param_dict)
        sorter.run()
        print(output_folder['value'])
        return sorter.get_result(), sorter.output_folders[0]



""" Model associated with user constructed pipeline of elements.

Supports the main UI using MVC pattern semantics.  This class proxies the
actual concatenation (pipeline) of SpikeInterface element space of extractors,
pre-processors, sorters, and curators.
"""

import PyQt5.QtCore as qc
import PyQt5.QtGui as qg
import PyQt5.QtWidgets as qw

import copy
import pkg_resources

from . import config as cfg

'''
from .config import ELEMENT_ROLE, main_window, status_bar, \
    STATUS_MSG_TIMEOUT, EXTRACTOR, SORTER
'''


class SpikePipelineModel(qc.QAbstractListModel):
    """Used by UI to display pipeline of elements in a decoupled fashion"""

    def __init__(self, element_model):
        """TBD."""
        super().__init__()

        # Underlying data structure proxied by model
        self._elements = []

        self._decorations = [None, None, None, None]
        fn = pkg_resources.resource_filename(
            'spikely.resources', 'EXTRACTOR.png')
        self._decorations[cfg.EXTRACTOR] = qg.QIcon(fn)

        fn = pkg_resources.resource_filename(
            'spikely.resources', 'PRE_PROCESSOR.png')
        self._decorations[cfg.PRE_PROCESSOR] = qg.QIcon(fn)

        fn = pkg_resources.resource_filename(
            'spikely.resources', 'SORTER.png')
        self._decorations[cfg.SORTER] = qg.QIcon(fn)

        fn = pkg_resources.resource_filename(
            'spikely.resources', 'CURATOR.png')
        self._decorations[cfg.CURATOR] = qg.QIcon(fn)

    # Methods sub-classed from QAbstractListModel
    def rowCount(self, parent):
        return len(self._elements)

    def data(self, mod_index, role=qc.Qt.DisplayRole):
        """ Retrieves data facet (role) from model based on positional index"""
        result = None

        if mod_index.isValid() and mod_index.row() < len(self._elements):
            element = self._elements[mod_index.row()]
            if role == qc.Qt.DisplayRole or role == qc.Qt.EditRole:
                result = element.name
            elif role == qc.Qt.DecorationRole:
                result = self._decorations[element.interface_id]
            elif role == cfg.ELEMENT_ROLE:
                result = element

        return result

    # Convenience methods used by class APIs
    def _has_instance(self, interface_id):
        for element in self._elements:
            if element.interface_id == interface_id:
                return True
        """Generator expression equivalent for future reference
        return sum(1 for ele in self._elements if
        ele.interface_id == interface_id)"""

    def _swap(self, list, pos1, pos2):
        list[pos1], list[pos2] = list[pos2], list[pos1]

    # Methods for other parts of Spikely to manipulate pipeline
    def run(self):

        bad_count = self._bad_param_count()
        if bad_count:
            qw.QMessageBox.warning(
                cfg.main_window, 'Run Failure',
                f'Missing {self._bad_param_count()} required ' +
                ('parameter' if bad_count == 1 else 'parameters'))
        else:
            """Call SpikeInterface APIs on elements in pipeline"""
            try:
                input_payload = None
                element_count = len(self._elements)

                for i in range(0, element_count):
                    next_element = self._elements[i+1] \
                        if i < (element_count - 1) else None
                    input_payload = self._elements[i].run(
                        input_payload, next_element
                    )

            except (KeyError, AttributeError):
                qw.QMessageBox.warning(
                    cfg.main_window, 'Run Failure',
                    'One or more invalid element parameter values.  Please '
                    'ensure all parameter values are set properly for all '
                    'elements in the pipeline.')
            except Exception as e:
                qw.QMessageBox.warning(
                    cfg.main_window, 'Run Failure', f'{e}')
            else:
                msg = 'Run successful' \
                    if element_count > 0 else 'Nothing to run'
                cfg.status_bar.showMessage(msg, cfg.STATUS_MSG_TIMEOUT)

    def clear(self):
        """Removes all elements from pipeline"""
        if len(self._elements):
            self.beginResetModel()
            self._elements.clear()
            self.endResetModel()
        else:
            cfg.status_bar.showMessage(
                'Nothing to clear', cfg.STATUS_MSG_TIMEOUT)

    def add_element(self, element):
        """ Adds element at top of stage associated w/ element interface_id"""
        # Only allow one Extractor or Sorter
        if (element.interface_id == cfg.EXTRACTOR or
                element.interface_id == cfg.SORTER):
            if self._has_instance(element.interface_id):
                cfg.status_bar.showMessage(
                    "Only one instance of that element type allowed",
                    cfg.STATUS_MSG_TIMEOUT)
                return
        # A bit hacky since it assumes order of interface_id constants
        i = 0
        while (i < len(self._elements) and
                element.interface_id >= self._elements[i].interface_id):
            i += 1
        self.beginInsertRows(qc.QModelIndex(), i, i)
        # Need a deep copy of element to support multi-instance element use
        self._elements.insert(i, copy.deepcopy(element))
        self.endInsertRows()

    def move_up(self, element):
        i = self._elements.index(element)
        # Elements confined to their stage
        if (i > 0 and self._elements[i].interface_id
                == self._elements[i-1].interface_id):
            self.beginMoveRows(qc.QModelIndex(), i, i, qc.QModelIndex(), i-1)
            self._swap(self._elements, i, i-1)
            self.endMoveRows()
        else:
            cfg.status_bar.showMessage(
                "Cannot move element any higher", cfg.STATUS_MSG_TIMEOUT)

    def move_down(self, element):
        i = self._elements.index(element)
        # Elements confined to their stage
        if (i < (len(self._elements) - 1) and
                self._elements[i].interface_id
                == self._elements[i+1].interface_id):
            # beginMoveRows behavior is fubar if move down from source to dest
            self.beginMoveRows(qc.QModelIndex(), i+1, i+1, qc.QModelIndex(), i)
            self._swap(self._elements, i, i+1)
            self.endMoveRows()
        else:
            cfg.status_bar.showMessage(
                "Cannot move element any lower", cfg.STATUS_MSG_TIMEOUT)

    def delete(self, element):
        index = self._elements.index(element)
        self.beginRemoveRows(qc.QModelIndex(), index, index)
        self._elements.pop(index)
        self.endRemoveRows()

    def _bad_param_count(self):
        count = 0
        for element in self._elements:
            for param in element.params:
                if 'value' not in param.keys():
                    count += 1
        return count



"""Creates an MVC view-control for configuring pipeline element properties.

The Configure Element view/control consists of widgets responsible for
viewing and editing the properties of elements (extractors, sorters, etc.).
"""

import PyQt5.QtWidgets as qw

from .config import PARAM_COL, VTYPE_COL, VALUE_COL


class ConfigureElementView(qw.QGroupBox):
    """GroupBox of widgets capable of editing element properties.

    Application instance of element model is updated everytime user selects an
    element in the pipeline view.  In turn, this view keys off of those
    element model changes to display properties associated with user selected
    element.
    """

    def __init__(self, pipeline_model, element_model):
        """Initialize parent, set object members, build UI region."""
        super().__init__("Configure Elements")
        self._pipeline_model = pipeline_model
        self._element_model = element_model

        self._init_ui()

    def _init_ui(self):
        """Build composite UI for element property configuration."""
        self.setLayout(qw.QHBoxLayout())

        cfg_table = qw.QTableView(self)

        # Key to refreshing widget contents
        cfg_table.setModel(self._element_model)

        cfg_table.verticalHeader().hide()
        cfg_table.setColumnWidth(PARAM_COL, 200)
        cfg_table.setColumnWidth(VTYPE_COL, 100)
        cfg_table.setColumnWidth(VALUE_COL, 200)
        cfg_table.horizontalHeader().setStretchLastSection(True)

        """
        cfg_table.horizontalHeader().setSectionResizeMode(
            1, qw.QHeaderView.Stretch)
        """

        self.layout().addWidget(cfg_table)



"""The view-control widget set for constructing the active pipeline.

The Construct Pipeline view-control consists of widgets responsible for
constructing the active pipeline by inserting, deleting, or moving elements
within the active pipeline.
"""

import PyQt5.QtWidgets as qw

from .extractor import Extractor
from .preprocessor import Preprocessor
from .sorter import Sorter
from .curator import Curator

from . import config as cfg

import spikeextractors as se
import spiketoolkit as st


class ConstructPipelineView(qw.QGroupBox):
    """QGroupBox containing the view-control widget set

    No public methods other than constructor.  All other activites
    of object are triggered by user interaction with sub widgets.
    """

    def __init__(self, pipeline_model, element_model):
        super().__init__("Construct Pipeline")

        self._available_elements = []
        self._get_available_elements()

        self._pipeline_model = pipeline_model
        self._pipeline_view = qw.QListView(self)
        self._element_model = element_model

        self._init_ui()

    def _init_ui(self):
        """Assembles the individual widgets into the widget-set.

        The ConstructPipelineView consists of three separate UI assemblies
        stacked top to bottom: element selection, pipeline element list view,
        and pipeline element manipulation controls (move up, delete, move down)
        """
        # Lay out view from top to bottom of group box
        self.setLayout(qw.QVBoxLayout())

        self.layout().addWidget(self._element_insertion())
        self.layout().addWidget(self._pipeline_list())
        self.layout().addWidget(self._pipeline_commands())

        """This funky bit of code is an example of how a class method
        with a specific signature can be bound to an instance of that class
        using a type index, in this case QModelIndex"""
        # treeView.clicked[QModelIndex].connect(self.clicked)

    def _element_insertion(self):
        """Select for and insert elements into pipeline."""

        ui_frame = qw.QFrame()
        ui_frame.setLayout(qw.QHBoxLayout())

        # Out of order declaration needed as forward reference
        ele_cbx = qw.QComboBox(self)

        """
        # Indicates user selection of combobox entry
        def _ele_cbx_activated(index):
            print('_ele_cbx activated')
        ele_cbx.activated.connect(_ele_cbx_activated)
        """

        stage_cbx = qw.QComboBox()
        ui_frame.layout().addWidget(stage_cbx)

        # Change ele_cbx contents when user makes stage_cbx selection
        def _stage_cbx_changed(index):
            ele_cbx.clear()
            for element in self._available_elements:
                # stage_cbx items store type ID associated w/ name
                if element.interface_id == stage_cbx.itemData(index):
                    # ele_cbx items store element object w/ name
                    ele_cbx.addItem(element.name, element)
        stage_cbx.currentIndexChanged.connect(_stage_cbx_changed)

        # Must come after currentIndexChanged.connect to invoke callback
        stage_cbx.addItem('Extractors', cfg.EXTRACTOR)
        stage_cbx.addItem('Pre-Processors', cfg.PRE_PROCESSOR)
        stage_cbx.addItem('Sorters', cfg.SORTER)
        stage_cbx.addItem('Curators', cfg.CURATOR)

        # Layout after stage_cbx, but declared first as fwd reference
        ui_frame.layout().addWidget(ele_cbx)

        add_button = qw.QPushButton("Add Element")

        def _add_element_clicked():
            # Prevents the addition of nu
            if ele_cbx.currentIndex() > -1:
                # Takes advantage of actual element reference stored in ele_cbx
                self._pipeline_model.add_element(ele_cbx.currentData())
        add_button.clicked.connect(_add_element_clicked)

        ui_frame.layout().addWidget(add_button)

        return ui_frame

    def _pipeline_list(self):
        # MVC in action - connect View (widget) to Model
        self._pipeline_view.setModel(self._pipeline_model)

        self._pipeline_view.setSelectionMode(
            qw.QAbstractItemView.SingleSelection)

        # Links element (ce_view) and pipeline (cp_view) views
        def list_selection_changed(selected, deselected):
            if selected.indexes():
                # Retrieve selected element from pipeline model
                element = self._get_selected_element()
                # Link selected element to element property editor
                self._element_model.element = element
            else:
                self._element_model.element = None
        self._pipeline_view.selectionModel().selectionChanged.connect(
            list_selection_changed)

        return self._pipeline_view

    def _pipeline_commands(self):
        ui_frame = qw.QFrame()
        ui_frame.setLayout(qw.QHBoxLayout())

        # Move Up element button and associated action
        mu_btn = qw.QPushButton("Move Up")
        ui_frame.layout().addWidget(mu_btn)

        def move_up_clicked():
            element = self._get_selected_element()
            if element is None:
                cfg.status_bar.showMessage(
                    "Nothing to move up", cfg.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.move_up(element)
        mu_btn.clicked.connect(move_up_clicked)

        # Move Down element button and associated action
        md_btn = qw.QPushButton("Move Down")
        ui_frame.layout().addWidget(md_btn)

        def move_down_clicked():
            element = self._get_selected_element()
            if element is None:
                cfg.status_bar.showMessage(
                    "Nothing to move down", cfg.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.move_down(element)
        md_btn.clicked.connect(move_down_clicked)

        # Delete element button and associated action
        de_btn = qw.QPushButton("Delete")
        ui_frame.layout().addWidget(de_btn)

        def delete_clicked():
            element = self._get_selected_element()
            if element is None:
                cfg.status_bar.showMessage(
                    "Nothing to delete", cfg.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.delete(element)
        de_btn.clicked.connect(delete_clicked)

        return ui_frame

    def _get_selected_element(self):
        """ Convenience function to retrieve selected element in pipe view"""
        element = None
        model = self._pipeline_view.selectionModel()
        if model.hasSelection():
            index = model.selectedIndexes()[0]
            element = self._pipeline_model.data(index, cfg.ELEMENT_ROLE)
        return element

    def _get_available_elements(self):
        extractor_list = se.extractorlist.installed_recording_extractor_list
        for extractor_class in extractor_list:
            self._available_elements.append(
                Extractor(extractor_class, cfg.EXTRACTOR))

        preprocessor_list = st.preprocessing.preprocessinglist. \
            installed_preprocessers_list
        for preprocessor_class in preprocessor_list:
            self._available_elements.append(
                Preprocessor(preprocessor_class, cfg.PRE_PROCESSOR)
            )

        sorter_list = st.sorters.installed_sorter_list
        for sorter_class in sorter_list:
            self._available_elements.append(
                Sorter(sorter_class, cfg.SORTER)
            )

        curator_list = st.curation.installed_curation_list
        for curator_class in curator_list:
            self._available_elements.append(
                Curator(curator_class, cfg.CURATOR)
            )



import copy


class SpikeElement:
    """Base class for SpikeInterface elements"""

    def __init__(self, interface_id, interface_class, interface_name):
        self._interface_id = interface_id
        self._interface_class = interface_class
        self._name = interface_name
        self._params = copy.deepcopy(interface_class.gui_params())

    @property
    def interface_id(self):
        return self._interface_id

    @property
    def interface_class(self):
        return self._interface_class

    @property
    def name(self):
        return self._name

    @property
    def params(self):
        return self._params

    def setup(self):
        pass

    def run(self, input_payload, next_element):
        pass



from .spike_element import SpikeElement


class Preprocessor(SpikeElement):
    """Preprocessor class"""

    def __init__(self, interface_class, interface_id):
        SpikeElement.__init__(self, interface_id, interface_class,
                              interface_class.preprocessor_name)

    def run(self, input_payload, next_element):
        params_dict = {}
        params_dict['recording'] = input_payload
        params = self._params
        for param in params:
            param_name = param['name']
            # param_type = param['type']
            # param_title = param['title']
            param_value = param['value']
            params_dict[param_name] = param_value
        pp = self._interface_class(**params_dict)
        return pp



"""Class definition of SpikeElement.

Implements the SpikeInterface elements responsible extracellular data
processing.

"""

import PyQt5.QtCore as qc
import PyQt5.QtGui as qg
import PyQt5.QtWidgets as qw

from . import config as cfg


class SpikeElementModel(qc.QAbstractTableModel):
    """Model representation of pipeline elements"""

    def __init__(self):
        self._element = None
        super().__init__()

    # Pythonic setters/getters
    @property
    def element(self):
        return self._element

    @element.setter
    def element(self, element):
        '''Ensures views are signaled on element changes'''
        self.beginResetModel()
        self._element = element
        self.endResetModel()

    # Methods sub-classed from QAbstractTableModel
    def rowCount(self, parent=qc.QModelIndex()):
        '''Count of parameters associated with an element instance'''
        return 0 if self._element is None else len(self._element.params)

    def columnCount(self, parent=qc.QModelIndex()):
        '''Number of display columns in parameter table view'''
        return 3

    def flags(self, mod_index):
        '''Sets UI policy for columns in table view for element'''
        flags = qc.QAbstractTableModel.flags(self, mod_index)
        col = mod_index.column()
        if col == cfg.PARAM_COL or col == cfg.VTYPE_COL:
            flags ^= qc.Qt.ItemIsSelectable
        elif col == cfg.VALUE_COL:
            flags |= qc.Qt.ItemIsEditable
        return flags

    def data(self, mod_index, role=qc.Qt.DisplayRole):
        '''Returns data for one row/param and one col/field at a time'''
        col, row = mod_index.column(), mod_index.row()
        param_dict = self._element.params[row]
        result = qc.QVariant()

        if role == qc.Qt.DisplayRole or role == qc.Qt.EditRole:
            if col == cfg.PARAM_COL:
                result = param_dict['name']
            elif col == cfg.VTYPE_COL:
                result = param_dict['type']
            elif col == cfg.VALUE_COL:
                if 'value' in param_dict.keys():
                    result = str(param_dict['value']).strip()
                    if not result and 'default' in param_dict.keys():
                        result = str(param_dict['default'])

        elif role == qc.Qt.ToolTipRole:
            if col == cfg.PARAM_COL and 'title' in param_dict.keys():
                result = param_dict['title']

        elif role == qc.Qt.BackgroundRole:
            if col == cfg.VALUE_COL:
                if ('value' not in param_dict.keys() and
                        'default' not in param_dict.keys()):
                    result = qg.QBrush(qg.QColor(255, 192, 192))

        return result

    def headerData(self, section, orientation, role):
        '''Sets properties for column headers in table view'''
        result = qc.QVariant()
        if (orientation == qc.Qt.Horizontal and
                (role == qc.Qt.DisplayRole or role == qc.Qt.EditRole)):
            result = ['Parameter', 'Type', 'Value'][section]

        return result

    def setData(self, mod_index, value, role=qc.Qt.EditRole):
        '''Takes data from table view and passes it to model'''
        row = mod_index.row()
        param_dict = self._element.params[row]
        success = True

        if role == qc.Qt.EditRole:
            '''
            This is a little tricky - if user enters valid value assign it to
            'value' in the param dictionary.  If user enters nothing assign
            'default' to 'value' if 'default' exists, otherwise delete 'value'
            from param dictionary.  Talk to Cole if you don't like this.
            '''
            if value.strip():
                success, cvt_value = self._convert_value(
                    param_dict['type'], value)
                if success:
                    param_dict['value'] = cvt_value
            else:
                if 'default' in param_dict.keys():
                    param_dict['value'] = param_dict['default']
                else:
                    param_dict.pop('value', None)

        return success

    def _convert_value(self, type_str, value):
        success, cvt_value = True, None
        try:
            if value == 'None':
                cvt_value = None
            elif type_str in ['str', 'path']:
                cvt_value = value
            elif type_str == 'int':
                cvt_value = int(value)
            elif type_str == 'float':
                cvt_value = float(value)
            elif type_str == 'bool':
                if value.lower() in ['true', 'yes']:
                    cvt_value = True
                elif value.lower() in ['false', 'no']:
                    cvt_value = False
                else:
                    raise TypeError(f'{value} is not a valid bool type')
            else:
                raise TypeError(f'{type_str} is not a Spikely supported type')
        except (TypeError, ValueError) as err:
            qw.QMessageBox.warning(
                cfg.main_window, 'Type Conversion Error', repr(err))
            success = False

        return success, cvt_value
